@@ -49,6 +49,12 @@ public:
   virtual ~DataSparse() override = default;
 
   double get(size_t row, size_t col) const override {
+    // Use permuted data for corrected impurity importance
+    size_t col_permuted = col;
+    if (col >= num_cols) {
+      col = getUnpermutedVarID(col);
+      row = getPermutedSampleID(row);
+    }
     return data.coeff(row, col);
   }
 
