@@ -1,52 +1,42 @@
 /*-------------------------------------------------------------------------------
- This file is part of Ranger.
+ This file is part of ranger.
 
- Ranger is free software: you can redistribute it and/or modify
- it under the terms of the GNU General Public License as published by
- the Free Software Foundation, either version 3 of the License, or
- (at your option) any later version.
+ Copyright (c) [2014-2018] [Marvin N. Wright]
 
- Ranger is distributed in the hope that it will be useful,
- but WITHOUT ANY WARRANTY; without even the implied warranty of
- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
- GNU General Public License for more details.
+ This software may be modified and distributed under the terms of the MIT license.
 
- You should have received a copy of the GNU General Public License
- along with Ranger. If not, see <http://www.gnu.org/licenses/>.
-
- Written by:
-
- Marvin N. Wright
- Institut für Medizinische Biometrie und Statistik
- Universität zu Lübeck
- Ratzeburger Allee 160
- 23562 Lübeck
- Germany
- 
- http://www.imbs-luebeck.de
+ Please note that the C++ core of ranger is distributed under MIT license and the
+ R package "ranger" under GPL3 license.
  #-------------------------------------------------------------------------------*/
 
 #ifndef TREEREGRESSION_H_
 #define TREEREGRESSION_H_
 
+#include <vector>
+
 #include "globals.h"
 #include "Tree.h"
 
+namespace ranger {
+
 class TreeRegression: public Tree {
 public:
-  TreeRegression();
+  TreeRegression() = default;
 
   // Create from loaded forest
   TreeRegression(std::vector<std::vector<size_t>>& child_nodeIDs, std::vector<size_t>& split_varIDs,
       std::vector<double>& split_values);
 
-  virtual ~TreeRegression();
+  TreeRegression(const TreeRegression&) = delete;
+  TreeRegression& operator=(const TreeRegression&) = delete;
 
-  void allocateMemory();
+  virtual ~TreeRegression() override = default;
+
+  void allocateMemory() override;
 
   double estimate(size_t nodeID);
   void computePermutationImportanceInternal(std::vector<std::vector<size_t>>* permutations);
-  void appendToFileInternal(std::ofstream& file);
+  void appendToFileInternal(std::ofstream& file) override;
 
   double getPrediction(size_t sampleID) const {
     size_t terminal_nodeID = prediction_terminal_nodeIDs[sampleID];
@@ -58,15 +48,18 @@ public:
   }
 
 private:
-  bool splitNodeInternal(size_t nodeID, std::vector<size_t>& possible_split_varIDs);
-  void createEmptyNodeInternal();
+  bool splitNodeInternal(size_t nodeID, std::vector<size_t>& possible_split_varIDs) override;
+  void createEmptyNodeInternal() override;
 
-  double computePredictionAccuracyInternal();
+  double computePredictionAccuracyInternal() override;
 
   // Called by splitNodeInternal(). Sets split_varIDs and split_values.
   bool findBestSplit(size_t nodeID, std::vector<size_t>& possible_split_varIDs);
   void findBestSplitValueSmallQ(size_t nodeID, size_t varID, double sum_node, size_t num_samples_node,
       double& best_value, size_t& best_varID, double& best_decrease);
+  void findBestSplitValueSmallQ(size_t nodeID, size_t varID, double sum_node, size_t num_samples_node,
+      double& best_value, size_t& best_varID, double& best_decrease, std::vector<double> possible_split_values,
+      std::vector<double>& sums_right, std::vector<size_t>& n_right);
   void findBestSplitValueLargeQ(size_t nodeID, size_t varID, double sum_node, size_t num_samples_node,
       double& best_value, size_t& best_varID, double& best_decrease);
   void findBestSplitValueUnordered(size_t nodeID, size_t varID, double sum_node, size_t num_samples_node,
@@ -77,6 +70,9 @@ private:
   bool findBestSplitExtraTrees(size_t nodeID, std::vector<size_t>& possible_split_varIDs);
   void findBestSplitValueExtraTrees(size_t nodeID, size_t varID, double sum_node, size_t num_samples_node,
       double& best_value, size_t& best_varID, double& best_decrease);
+  void findBestSplitValueExtraTrees(size_t nodeID, size_t varID, double sum_node, size_t num_samples_node,
+      double& best_value, size_t& best_varID, double& best_decrease, std::vector<double> possible_split_values,
+      std::vector<double>& sums_right, std::vector<size_t>& n_right);
   void findBestSplitValueExtraTreesUnordered(size_t nodeID, size_t varID, double sum_node, size_t num_samples_node,
       double& best_value, size_t& best_varID, double& best_decrease);
 
@@ -84,19 +80,17 @@ private:
 
   double computePredictionMSE();
 
-  void cleanUpInternal() {
-    if (counter != 0) {
-      delete[] counter;
-    }
-    if (sums != 0) {
-      delete[] sums;
-    }
+  void cleanUpInternal() override {
+    counter.clear();
+    counter.shrink_to_fit();
+    sums.clear();
+    sums.shrink_to_fit();
   }
 
-  size_t* counter;
-  double* sums;
-
-  DISALLOW_COPY_AND_ASSIGN(TreeRegression);
+  std::vector<size_t> counter;
+  std::vector<double> sums;
 };
 
+} // namespace ranger
+
 #endif /* TREEREGRESSION_H_ */
